#include <NTL/ZZ_pXFactoring.h>
#include <flint/fmpz.h>
#include <flint/fmpz_mod_poly.h>
#include <cstdio>

NTL_CLIENT

struct FlintZZ {
   fmpz_t value;

   explicit
   FlintZZ(const ZZ& a)
   {
      long n = NumBits(a);
   
      fmpz_init(value);
   
      for (long i = 0; i < n; i++)
         if (bit(a, i)) fmpz_setbit(value, i);
   
      if (a < 0)
         fmpz_neg(value, value);

   }

   ~FlintZZ() { fmpz_clear(value); }

};

struct FlintModulus {
   fmpz_mod_ctx_t value;

   explicit FlintModulus(const ZZ& p)
   {
      FlintZZ f_p(p);
      fmpz_mod_ctx_init(value, f_p.value);
   }

   ~FlintModulus()
   {
      fmpz_mod_ctx_clear(value);
   }
};



struct FlintZZ_pX {
   fmpz_mod_poly_t value;
   const FlintModulus& mod;

   FlintZZ_pX(const ZZ_pX& a, const FlintModulus& _mod) : mod(_mod)
   {
      long da = deg(a);
      fmpz_mod_poly_init2(value, da+1, mod.value);

      for (long i = 0; i <= da; i++) {
         FlintZZ f_c(rep(a[i]));
         fmpz_mod_poly_set_coeff_fmpz(value, i, f_c.value, mod.value);
      }
   }

   explicit
   FlintZZ_pX(const FlintModulus& _mod) : mod(_mod)
   {
      fmpz_mod_poly_init(value, mod.value);
   }

   ~FlintZZ_pX() { fmpz_mod_poly_clear(value, mod.value); }
};

int main()
{
   setbuf(stdout, NULL);

   for (long l = 256; l <= 4096; l *= 2) {
      // for (long n = 256; n <= 16384; n *= 2) {
      for (long idx = 0; idx < 9; idx ++) {
         long n  = 256*(1L << idx/2);
         if (idx & 1) n += n/2;
         SetSeed((ZZ(l) << 64) + ZZ(n));

	 ZZ p;

	 RandomPrime(p, l);
	 ZZ_p::init(p);

         ZZ_pX f;
      
         random(f, n);
         SetCoeff(f, n);

         ZZ_pX g, h, res;
         random(g, n);
         random(h, n);
      

         double t;

         CompMod(res, g, h, f);

         long iter = 1;
         do {
            t = GetTime();
            for (long i = 0; i < iter; i++) CompMod(res, g, h, f);
            t = GetTime() - t;
            iter *= 2;
         } while (t < 3);
         iter /= 2;

         double NTLTime = t;

         FlintModulus mod(ZZ_p::modulus());
      
         FlintZZ_pX f_f(f,mod);
         FlintZZ_pX f_g(g,mod);
         FlintZZ_pX f_h(h,mod);
         FlintZZ_pX f_res(res,mod);
      
         t = GetTime();
         for (long i = 0; i < iter; i++) 
            fmpz_mod_poly_compose_mod(f_res.value, f_g.value, f_h.value, f_f.value, mod.value);
         t = GetTime()-t;
         double FlintTime = t;
      
         printf("%8.2f", FlintTime/NTLTime);
      }

      printf("\n");
   }
}


