#include <NTL/mat_lzz_p.h>
#include <flint/nmod_mat.h>
#include <cstdio>

NTL_CLIENT


struct Flint_mat_zz_p {
   nmod_mat_t value;

   explicit
   Flint_mat_zz_p(const mat_zz_p& a)
   {
      long n = a.NumRows();
      long m = a.NumCols();
      nmod_mat_init(value, n, m, zz_p::modulus());

      for (long i = 0; i < n; i++) {
         for (long j = 0; j < m; j++) {
            (*nmod_mat_entry_ptr(value, i, j)) = rep(a[i][j]);
         }
      }
   }

   ~Flint_mat_zz_p() { nmod_mat_clear(value); }
};




void FillRandom(Mat<zz_p>& A)
{
   long n = A.NumRows();
   long m = A.NumCols();
   for (long i = 0; i < n; i++)
      for (long j = 0; j < m; j++)
         random(A[i][j]);
}


void FillRandom(Vec<zz_p>& A)
{
   long n = A.length();
   for (long i = 0; i < n; i++)
      random(A[i]);
}

int main()
{
   setbuf(stdout, NULL);

   for (long l = 5; l <= 60; l += 5) {
      // n = 512 768 1024 1536 2048 3072 4096 
      for (long idx = 0; idx < 9; idx ++) {
         long n  = 512*(1L << idx/2);
         if (idx & 1) n += n/2;

	 SetSeed((ZZ(l) << 64) + ZZ(n));

	 long p;

	 p = RandomPrime_long(l);
	 zz_p::init(p);


	 mat_zz_p A, B, R;

	 B.SetDims(n, n);
	 clear(B);
	 for (long i = 0; i < n/2; i++) FillRandom(B[i]);
	 
	 R.SetDims(n, n);
	 FillRandom(R);

	 mul(A, B, R);
	 // A is (roughly) a random n/2-rank matrix of dimension nxn

	 B.kill();
	 R.kill();

	 mat_zz_p K;


	 double t;

	 long iter = 1;
	 do {
	    t = GetTime();
	    for (long i = 0; i < iter; i++) kernel(K, A);
	    t = GetTime() - t;
	    iter *= 2;
	 } while (t < 3);
	 iter /= 2;

	 double NTLTime = t;

	 mat_zz_p At;
	 transpose(At, A);
	 Flint_mat_zz_p fAt(At), fK(At);

	 t = GetTime();
	 for (long i = 0; i < iter; i++) nmod_mat_nullspace(fK.value, fAt.value);
	 t = GetTime()-t;
	 double FlintTime = t;

	 printf("%8.2f", FlintTime/NTLTime);
      }

      printf("\n");
   }
}


